unit FIB_dataset;

interface
{$I Version.inc}
uses
  Windows, Types, Classes, SysUtils, FmtBcd, StrUtils, Forms, Dialogs,
  Variants, Math, DateUtils, DB, DBConsts, RTLConsts,
  FIB_header, FIB_messages, FIB_params, FIB_SQLQuery, FIB_access,
  FIB_array, FIB_blob,
  CommonStuff;

type
  TRecordPosition = (rpFirst, rpNext, rpCurrent, rpPrior, rpLast);
  TCachedUpdDefHandling = (cuApply, cuCancel);
  TDataAccessMethod = (daDynamicAccess, daPartialFetch, daStaticAccess);
  TPartialFetchRefreshMethod = (pfrmRefetchStoredRecords, pfrmReleaseStoredRecords);
  TSortFieldsNaming = (sfnTDataset, sfnSQLServer);
  TSortMethod = (smSortOnServer, smReplaceIndexes, smPreserveIndexes, smMaintainAllIndexes);
  TSortState = (ssNotSorted, ssOrderBy, ssIndexFields);

  TRowIDFieldDef = packed
  record
    DataType: TFieldType;
    FetchedFieldIdx: Word;
    StoredFieldIdx1: Word;
    Offset: Word;
    Size: Byte;
  end;

  PRecordInfo = ^TRecordInfo;
  TRecordInfo = packed
  record
    RecStatus: Byte;
    RecordID: Pointer;
    BookmarkFlag: TBookmarkFlag;
  end;

  TFieldInfo = packed
  record
    Field: TField;
    Size: Word;
    Offset: Word;
  end;

  TSQLStatement = class(TPersistent)
  private
  protected
    FLoading: Boolean;
    procedure SetDataSetSQL;
    procedure AssignTo(Dest: TPersistent); override;    
    function AddWGOclause(ClauseID: Byte; aClause: string): string;
  public
    constructor Create(AOwner: TDataSet);
    destructor Destroy; override;
  published
    property SelectFrom: TStrings read FSelectFrom write SetSelectFrom;
    property Where: TStrings read FWhere write SetWhere;
    property GroupBy: string read FGroupBy write SetGroupBy;
    property Plan: TStrings read FPlan write SetPlan;
    property OrderBy: string read FOrderBy write SetOrderBy;
  end;

  TTableOptions = class(TPersistent)
  private
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create(ADataSet: TDataSet); virtual;
  published
    property CachedUpdates: Boolean read GetCachedUpdates write SetCachedUpdates;
    property CachedUpdDefHandling: TCachedUpdDefHandling read FCachedUpdDefHandling
             write FCachedUpdDefHandling;
    property DataAccessMethod: TDataAccessMethod read GetDataAccessMethod
             write SetDataAccessMethod;
    property PartialFetchRecount: Byte read GetPartialFetchRecount
             write SetPartialFetchRecount;
    property PartialFetchRefreshMethod: TPartialFetchRefreshMethod
             read GetPartialFetchRefreshMethod
             write SetPartialFetchRefreshMethod
             default pfrmRefetchStoredRecords;
  end;


  TFIB_Dataset = class(TDataset)
  private
  protected
    procedure Loaded; override;
    procedure SetActive(Value: Boolean); override;
    procedure Allocate_BLOB_ARRAY_Pointers(Buffer: PChar); virtual;
    procedure Free_BLOB_ARRAY_Pointers(Buffer: PChar); virtual;
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;

    function GetDataSource: TDataSource; override;

    function GetFieldClass(FieldType: TFieldType): TFieldClass; override;

    procedure OpenCursor(InfoQuery: Boolean = False); override;
    procedure CloseCursor; override;

    procedure InternalInitFieldDefs; override;
    procedure InternalOpen; override;
    procedure InternalClose; override;

    procedure InternalFirst; override;
    procedure InternalLast; override;

    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalInsert; override;
    procedure InternalEdit; override;
    procedure InternalCancel; override;
    procedure InternalPost; override;
    procedure InternalDelete; override;
    procedure InternalRefresh; override;

    procedure InternalHandleException; override;

    function IsCursorOpen: Boolean; override;
    function GetCanModify: Boolean; override;
    procedure ClearCalcFields(Buffer: PChar); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    procedure SetCurrentRecord(Index: Integer); override;

    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;

    procedure ValidateInsert(AComponent: TComponent); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DataEvent(Event: TDataEvent; Info: Longint); override;

    procedure SetFiltered(Value: Boolean); override;
    procedure SetFilterText(const Value: string); override;

    property IsTable: Boolean read FisTable write SetIsTable;
    property StoredRecordCount: Integer read GetStoredRecordCount;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    
    function GetCurrentRecord(Buffer: PChar): Boolean; override;
    function UpdateStatus: TUpdateStatus; override;
    
    procedure ApplyUpdates;
    procedure CancelUpdates;
    
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function BookmarkValid(Bookmark: TBookmark): Boolean; override;
    function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;

    function FindField(const FieldName: string): TField;
    function FieldByName(const FieldName: string): TField;
    function FieldByNumber(const FieldNo: Integer): TField;

    function LocateInTable(const KeyFields: string; const KeyValues: Variant;
                           Options: TLocateOptions): Boolean;
    function Locate(const KeyFields: string; const KeyValues: Variant;
                    Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant;
                    const ResultFields: string): Variant; override;
    function FindKey(const KeyValues: array of const; Options: TLocateOptions): Boolean;
    procedure FindNearest(const KeyValues: array of const; Options: TLocateOptions = []);
    procedure LocateNearest(const KeyFields: string; const KeyValues: array of const;
                            Options: TLocateOptions = []); overload;
    procedure LocateNearest(const KeyFields: string; const KeyValues: Variant;
                            Options: TLocateOptions = []); overload;

    procedure ChangeSorting(SortFieldNames: AnsiString; SortFieldsNaming: TSortFieldsNaming);

    property IndexFieldNames: string read FIndexFieldNames write SetIndexFieldNames;
    property RowIDFields: string read FRowIDFields;
    property RowIDFieldByName[FieldName: string]: Variant read GetRowIDFieldValueByName;
    property BlobArrayID[FieldName: String]: TGDS_QUAD read GetBlobArrayID;
  published
    property AutoDBConnect: Boolean read FAutoDBConnect write FAutoDBConnect default True;
    property Database: TFIB_Database read FDatabase write SetDatabase;
    property Transaction: TFIB_Transaction read FTransaction write SetTransaction;
    property SQLStatement: TSQLStatement read FSQLStatement write SetSQLStatement;
    property Params: TFIB_Params read FParams write SetParams;
    property ProcName: string read FProcName write SetProcName;
    property TableName: string read FTableName write SetTableName;
    property TableOptions: TTableOptions read FTableOptions write SetTableOptions;
    property FetchAllOnOpen: Boolean read FFetchAllOnOpen write SetFetchAllOnOpen;
    property UniDirectional: Boolean read UniDirectionalGet
             write UniDirectionalSet default False;
    property LinkedFields: AnsiString read FLinkedFieldNames write FLinkedFieldNames;
    property MasterFields: string read GetMasterFields write SetMasterFields;
    property MasterSource: TDataSource read GetDataSource write SetMasterSource;
    property KeepCursorPosOnRefresh: Boolean read FKeepCursorPosOnRefresh write FKeepCursorPosOnRefresh;
    property SortMethod: TSortMethod read FSortMethod write SetSortMethod;
    property ShowSQLScreenCursor: Boolean read FShowSQLScreenCursor
             write FShowSQLScreenCursor default True;
    property Active;
    property AutoCalcFields;
    property Filter;
    property Filtered;
    { Events }
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeRefresh;
    property AfterRefresh;
    property BeforeScroll;
    property AfterScroll;
    property OnNewRecord;
    property OnCalcFields;
    property OnEditError;
    property OnPostError;
    property OnDeleteError;
    property OnFilterRecord;
  end;

  TFIB_BlobStream = class(TMemoryStream)
  private
  protected
    function Realloc(var NewCapacity: Longint): Pointer; override;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode); reintroduce; overload;
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
  end;

{$IFNDEF Delphi7}
  TFIBLargeIntField = class(TLargeIntField)
  protected
    procedure SetVarValue(const Value: Variant); override;
  end;
{$ENDIF}

  TFIBArrayField = class(TFIBCustomArrayField)
  protected
    function GetDatabase: PFIB_Database; override;
    function GetTRHandle: PISC_TR_HANDLE; override;
  end;


implementation

end.
